var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
import mapSeries from "p-each-series";
import PQueue from "p-queue";
import { Log, JSON_ENCODING, } from "@dao-xyz/peerbit-log";
import { Entry } from "@dao-xyz/peerbit-log";
import { variant, option, field, vec } from "@dao-xyz/borsh";
import { serialize, deserialize } from "@dao-xyz/borsh";
import { Snapshot } from "./snapshot.js";
import { AccessError, } from "@dao-xyz/peerbit-crypto";
import { waitForAsync } from "@dao-xyz/peerbit-time";
import { logger as loggerFn } from "@dao-xyz/peerbit-logger";
import path from "path-browserify";
import { v4 as uuid } from "uuid";
import { join } from "./replicator.js";
import { createBlock, getBlockValue } from "@dao-xyz/libp2p-direct-block";
export const logger = loggerFn({ module: "store" });
export class CachedValue {
}
let CID = class CID {
    hash;
    constructor(opts) {
        if (opts) {
            this.hash = opts.hash;
        }
    }
};
__decorate([
    field({ type: "string" }),
    __metadata("design:type", String)
], CID.prototype, "hash", void 0);
CID = __decorate([
    variant(0),
    __metadata("design:paramtypes", [Object])
], CID);
export { CID };
let CachePath = class CachePath {
    path;
    constructor(path) {
        this.path = path;
    }
};
__decorate([
    field({ type: "string" }),
    __metadata("design:type", String)
], CachePath.prototype, "path", void 0);
CachePath = __decorate([
    variant(0),
    __metadata("design:paramtypes", [String])
], CachePath);
export { CachePath };
let UnsfinishedReplication = class UnsfinishedReplication {
    hashes;
    constructor(opts) {
        if (opts) {
            this.hashes = opts.hashes;
        }
    }
};
__decorate([
    field({ type: vec("string") }),
    __metadata("design:type", Array)
], UnsfinishedReplication.prototype, "hashes", void 0);
UnsfinishedReplication = __decorate([
    variant(0),
    __metadata("design:paramtypes", [Object])
], UnsfinishedReplication);
export { UnsfinishedReplication };
let HeadsCache = class HeadsCache {
    heads;
    last;
    counter;
    constructor(heads, counter, last) {
        this.heads = heads;
        this.last = last;
        this.counter = counter;
    }
};
__decorate([
    field({ type: vec("string") }),
    __metadata("design:type", Array)
], HeadsCache.prototype, "heads", void 0);
__decorate([
    field({ type: option("string") }),
    __metadata("design:type", String)
], HeadsCache.prototype, "last", void 0);
__decorate([
    field({ type: "u64" }),
    __metadata("design:type", BigInt)
], HeadsCache.prototype, "counter", void 0);
HeadsCache = __decorate([
    variant(0),
    __metadata("design:paramtypes", [Array, BigInt, String])
], HeadsCache);
export { HeadsCache };
export const DefaultOptions = {
    replicationConcurrency: 32,
    cacheId: "id",
};
let Store = class Store {
    _storeIndex; // how to ensure unqiueness
    _canAppend;
    _onUpdate;
    _onUpdateOption;
    // An access controller that is note part of the store manifest, usefull for circular store -> access controller -> store structures
    _options;
    identity;
    _headsPathCounter = 0;
    headsPath;
    _lastHeadsPath;
    _lastHeadsCount = 0n;
    removedHeadsPath;
    _lastRemovedHeadsPath;
    _lastRemovedHeadsCount = 0n;
    snapshotPath;
    initialized;
    encoding = JSON_ENCODING;
    _store;
    _cache;
    _oplog;
    _cacheWriteQueue;
    _key;
    _saveFile;
    _loadFile;
    _loaded = false;
    constructor(properties) {
        if (properties) {
            this._storeIndex = properties?.storeIndex;
        }
    }
    setup(properties) {
        this.encoding = properties.encoding;
        this.onUpdate = properties.onUpdate;
        this.canAppend = properties.canAppend;
    }
    async init(store, identity, options) {
        if (this.initialized) {
            throw new Error("Already initialized");
        }
        this._saveFile = async (file) => store.put(await createBlock(file, "raw"));
        this._loadFile = async (file) => {
            const block = await store.get(file);
            if (block) {
                return getBlockValue(block);
            }
            return undefined;
        };
        // Set ipfs since we are to save the store
        this._store = store;
        // Set the options (we will use the topic property after thiis)
        const opts = { ...DefaultOptions, ...options };
        this._options = opts;
        // Cache
        this._cache = await this._options.resolveCache(this);
        await this._cache.open();
        // Create IDs, names and paths
        this.identity = identity;
        this._onUpdateOption = options.onUpdate;
        this.headsPath = path.join(options.cacheId, this.id, "_heads");
        this.removedHeadsPath = path.join(options.cacheId, this.id, "_heads_removed");
        await this.loadLastHeadsPath();
        this.snapshotPath = path.join(options.cacheId, this.id, "snapshot");
        // Create the operations log
        this._oplog = new Log(this._store, identity, this.logOptions);
        // addOperation and log-joins queue. Adding ops and joins to the queue
        // makes sure they get processed sequentially to avoid race conditions
        this._cacheWriteQueue = new PQueue({ concurrency: 1 });
        if (this._options.onOpen) {
            await this._options.onOpen(this);
        }
        this.initialized = true;
        return this;
    }
    /* private _updateCacheHeadsJob: any;
    private _updateCacheChange: {
        added: (Entry<T> | string)[];
        removed: (Entry<T> | string)[];
    }
    updateCacheHeads(change: {
        added: (Entry<T> | string)[];
        removed: (Entry<T> | string)[];
    },
        reset?: boolean) {
        clearTimeout(this._updateCacheHeadsJob)
        this._updateCacheHeadsJob = setTimeout(() => {
            this._cacheWriteQueue.add(() =>)
        }, 100)
    } */
    async _updateCachedHeads(change, reset) {
        if (typeof reset !== "boolean") {
            // Only reset all heads if loaded once, since we don't want too loose track of unloaded heads
            if (this._loaded &&
                this.oplog.headsIndex.index.size <= change.added.length) {
                let addedIsAllHeads = true;
                for (const entry of change.added) {
                    const hash = typeof entry === "string" ? entry : entry.hash;
                    if (!this.oplog.headsIndex.has(hash)) {
                        addedIsAllHeads = false;
                    }
                }
                reset = addedIsAllHeads;
            }
            else {
                // added size < head size, meaning we have not rewritten all heads
                reset = false;
            }
        }
        // If 'reset' then dont keep references to old heads caches, assume new cache will fully describe all heads
        const updateHashes = async (headsPath, lastPath, lastCounter, hashes) => {
            const newHeadsPath = path.join(headsPath, String(this._headsPathCounter), uuid());
            const counter = lastCounter + BigInt(hashes.length);
            await Promise.all([
                this._cache.set(headsPath, serialize(new CachePath(newHeadsPath.toString()))),
                this._cache.set(newHeadsPath, serialize(new HeadsCache(hashes, counter, lastPath))),
            ]);
            return { counter, newPath: newHeadsPath };
        };
        // TODO dont delete old before saving new
        if (reset) {
            const paths = [
                path.join(this.headsPath, String(this._headsPathCounter)),
                path.join(this.removedHeadsPath, String(this._headsPathCounter)),
            ];
            for (const p of paths) {
                await this._cache.deleteByPrefix(p + "/");
            }
            this._lastHeadsPath = undefined;
            this._lastRemovedHeadsPath = undefined;
            this._lastHeadsCount = 0n;
            this._lastRemovedHeadsCount = 0n;
            this._headsPathCounter += 1;
        }
        if (change.added.length > 0) {
            const update = await updateHashes(this.headsPath, this._lastHeadsPath, this._lastHeadsCount, change.added.map((x) => (typeof x === "string" ? x : x.hash)));
            this._lastHeadsPath = update.newPath;
            this._lastHeadsCount = update.counter;
        }
        if (this._lastHeadsPath) {
            // only add removed heads if we actually have added heads, else these are pointless
            if (change.removed.length > 0) {
                const update = await updateHashes(this.removedHeadsPath, this._lastRemovedHeadsPath, this._lastRemovedHeadsCount, change.removed.map((x) => (typeof x === "string" ? x : x.hash)));
                this._lastRemovedHeadsPath = update.newPath;
                this._lastRemovedHeadsCount = update.counter;
                if (update.counter > 0n &&
                    2n * update.counter >= this._lastHeadsCount) {
                    const resetToHeads = await this.getCachedHeads(this._lastHeadsPath, this._lastRemovedHeadsPath);
                    await this._updateCachedHeads({ added: resetToHeads, removed: [] }, true);
                }
            }
        }
    }
    async idle() {
        // Wait for the operations queue to finish processing
        // to make sure everything that all operations that have
        // been queued will be written to disk
        await this._cacheWriteQueue?.onIdle();
        await this._cache?.idle();
    }
    async getCachedHeads(lastHeadsPath = this._lastHeadsPath, lastRemovedHeadsPath = this._lastRemovedHeadsPath) {
        if (!this._cache) {
            return [];
        }
        const getHashes = async (start, filter) => {
            const result = [];
            let next = start;
            while (next) {
                const cache = await this._cache
                    .get(next)
                    .then((bytes) => bytes && deserialize(bytes, HeadsCache));
                next = cache?.last;
                cache?.heads.forEach((head) => {
                    if (filter && filter.has(head)) {
                        return;
                    }
                    result.push(head);
                });
            }
            return result;
        };
        const removedHeads = new Set(await getHashes(lastRemovedHeadsPath));
        const heads = await getHashes(lastHeadsPath, removedHeads);
        return heads; // Saved heads - removed heads
    }
    get id() {
        if (typeof this._storeIndex !== "number") {
            throw new Error("Store index not set");
        }
        return this._storeIndex.toString();
    }
    get oplog() {
        return this._oplog;
    }
    get key() {
        return this._key;
    }
    get store() {
        return this._store;
    }
    get options() {
        return this._options;
    }
    get logOptions() {
        return {
            logId: this.id,
            encryption: this._options.encryption,
            encoding: this.encoding,
            sortFn: this._options.sortFn,
            trim: this._options.trim && {
                // I can trim if I am not a replicator of an entry
                ...this._options.trim,
                filter: this.options.replicator && {
                    canTrim: async (gid) => !(await this.options.replicator(gid)),
                    cacheId: this.options.replicatorsCacheId,
                },
            },
        };
    }
    get cache() {
        return this._cache;
    }
    setIdentity(identity) {
        this.identity = identity;
        this._oplog.setIdentity(identity);
    }
    set canAppend(canAppend) {
        this._canAppend = canAppend;
    }
    get canAppend() {
        return this._canAppend;
    }
    set onUpdate(onUpdate) {
        this._onUpdate = onUpdate;
    }
    get closed() {
        return !this._oplog;
    }
    async close() {
        if (!this.initialized) {
            return;
        }
        if (this._options.onClose) {
            await this._options.onClose(this);
        }
        await this.idle();
        await this._cache.close();
        this._loaded = false;
        this._oplog = null;
        this._lastHeadsPath = undefined;
        this._lastRemovedHeadsPath = undefined;
        this._lastRemovedHeadsCount = 0n;
        this._lastHeadsCount = 0n;
        // Database is now closed
        return Promise.resolve();
    }
    /**
     * Drops a database and removes local data
     */
    async drop() {
        if (!this._oplog && !this._cache) {
            return; // already dropped
        }
        if (this._options.onDrop) {
            await this._options.onDrop(this);
        }
        if (this._cache.status !== "open") {
            await this._cache.open();
        }
        await this._cache.del(this.headsPath);
        await this._cache.del(this.snapshotPath);
        await this.close();
        // Reset
        // TODO fix types
        this._oplog = undefined;
        this._cache = undefined;
        this.initialized = false; // call this last because (close() expect initialized to be able to function)
    }
    async loadHeads() {
        if (!this.initialized) {
            throw new Error("Store needs to be initialized before loaded");
        }
        if (this._cache.status !== "open") {
            await this._cache.open();
        }
        await this.loadLastHeadsPath();
        const heads = await this.getCachedHeads(this._lastHeadsPath, this._lastRemovedHeadsPath);
        return heads;
    }
    async load(opts = {}) {
        if (!this.initialized) {
            throw new Error("Store needs to be initialized before loaded");
        }
        if (this._cache.status !== "open") {
            await this._cache.open();
        }
        if (this._options.onLoad) {
            await this._options.onLoad(this);
        }
        const heads = await this.loadHeads();
        // Load the log
        if (opts.heads) {
            const entries = await Promise.all(heads.map((x) => Entry.fromMultihash(this._store, x, {
                timeout: opts?.fetchEntryTimeout,
                replicate: true,
            })));
            this._oplog = new Log(this._store, this.identity, {
                ...this.logOptions,
                concurrency: this._options.replicationConcurrency,
            });
            await this._oplog.reset(entries);
        }
        else {
            const amount = opts.amount;
            if (amount != null && amount >= 0 && amount < heads.length) {
                throw new Error("You are not loading all heads, this will lead to unexpected behaviours on write. Please load at least load: " +
                    amount +
                    " entries");
            }
            this._oplog = await Log.fromEntryHash(this._store, this.identity, heads, {
                ...this.logOptions,
                length: amount ?? -1,
                timeout: opts?.fetchEntryTimeout,
                onFetched: this._onLoadProgress.bind(this),
                concurrency: this._options.replicationConcurrency,
            });
        }
        // Update the index
        if (heads.length > 0) {
            await this._updateIndex({
                added: await this._oplog.toArray(),
                removed: [],
            });
        }
        await this._updateCachedHeads({ added: [...this._oplog.headsIndex.index.values()], removed: [] }, true);
        this._loaded = true;
        this._options.onReady && this._options.onReady(this);
    }
    async loadLastHeadsPath() {
        this._lastHeadsPath = await this._cache
            .get(this.headsPath)
            .then((bytes) => bytes && deserialize(bytes, CachePath).path);
        this._lastRemovedHeadsPath = await this._cache
            .get(this.removedHeadsPath)
            .then((bytes) => bytes && deserialize(bytes, CachePath).path);
        this._lastHeadsCount = this._lastHeadsPath
            ? await this.getCachedHeadsCount(this._lastHeadsPath)
            : 0n;
        this._lastRemovedHeadsCount = this._lastRemovedHeadsPath
            ? await this.getCachedHeadsCount(this._lastRemovedHeadsPath)
            : 0n;
    }
    async getCachedHeadsCount(headPath) {
        if (!headPath) {
            return 0n;
        }
        return ((await this._cache
            .get(headPath)
            .then((bytes) => bytes && deserialize(bytes, HeadsCache)))?.counter || 0n);
    }
    async waitForHeads() {
        if (!this._loaded) {
            await this._cacheWriteQueue.add(async () => {
                if (this._loaded) {
                    return;
                }
                return this.load({ heads: true });
            });
        }
    }
    async addOperation(data, options) {
        if (this.closed) {
            throw new Error("Store is closed");
        }
        await this.waitForHeads();
        const change = await this._oplog.append(data, {
            nexts: options?.nexts,
            reciever: options?.reciever,
            canAppend: options?.skipCanAppendCheck ? undefined : this.canAppend,
            identity: options?.identity,
            signers: options?.signers,
        });
        logger.debug("Appended entry with hash: " + change.entry.hash);
        const changes = {
            added: [change.entry],
            removed: change.removed,
        };
        await Promise.all([
            this._cacheWriteQueue.add(() => this._updateCachedHeads(changes)),
            this._updateIndex(changes),
        ]);
        this._options.onWrite && this._options.onWrite(this, change.entry);
        return change;
    }
    async removeEntry(entry, options) {
        await this.waitForHeads();
        const entries = Array.isArray(entry) ? entry : [entry];
        if (entries.length === 0) {
            return {
                added: [],
                removed: [],
            };
        }
        if (options?.recursively) {
            await this.oplog.deleteRecursively(entry);
        }
        else {
            for (const entry of entries) {
                await this.oplog.delete(entry);
            }
        }
    }
    async removeOperation(entry, options) {
        await this.removeEntry(entry, options);
        const change = {
            added: [],
            removed: Array.isArray(entry) ? entry : [entry],
        };
        await Promise.all([
            this._updateCachedHeads(change),
            this._updateIndex(change),
        ]);
        return change;
    }
    /**
     *
     * @param entries
     * @returns change
     */
    async sync(entries, options = {}) {
        logger.debug(`Sync request #${entries.length}`);
        if (entries.length === 0) {
            return false;
        }
        const handle = async (headToHandle) => {
            const allEntries = headToHandle instanceof Entry
                ? [headToHandle]
                : [headToHandle.entry, ...headToHandle.references];
            await Promise.all(allEntries.map((h) => h.init({
                encoding: this.oplog.encoding,
                encryption: this.oplog.encryption,
            })));
            const entry = headToHandle instanceof Entry ? headToHandle : headToHandle.entry;
            this._options.onReplicationQueued &&
                this._options.onReplicationQueued(this, entry);
            const canAppend = options?.canAppend || this.canAppend;
            if (canAppend && !(await canAppend(entry))) {
                logger.debug("Not allowd to append head " + entry.hash);
                return Promise.resolve(null);
            }
            await Promise.all(allEntries.map(async (head) => {
                const headHash = head.hash;
                head.hash = undefined;
                try {
                    const hash = await Entry.toMultihash(this._store, head);
                    head.hash = headHash;
                    if (head.hash === undefined) {
                        head.hash = hash; // can happen if you sync entries that you load directly from ipfs
                    }
                    else if (hash !== head.hash) {
                        logger.error("Head hash didn't match the contents");
                        throw new Error("Head hash didn't match the contents");
                    }
                }
                catch (error) {
                    logger.error(error);
                    throw error;
                }
            }));
            return headToHandle;
        };
        const hash = (entry) => {
            if (entry instanceof Entry) {
                return entry.hash;
            }
            else if (typeof entry === "string") {
                return entry;
            }
            return entry.entry.hash;
        };
        const newEntries = [];
        for (const entry of entries) {
            const h = hash(entry);
            if (h && this.oplog.has(h)) {
                continue;
            }
            newEntries.push(typeof entry === "string"
                ? await Entry.fromMultihash(this._store, entry)
                : entry);
        }
        if (newEntries.length === 0) {
            return false;
        }
        const saved = await mapSeries(newEntries, handle);
        const change = await join(saved, this._oplog, {
            concurrency: this._options.replicationConcurrency,
            onFetched: (entry) => this._options.onReplicationFetch &&
                this._options.onReplicationFetch(this, entry),
        });
        await Promise.all([
            this._updateCachedHeads(change),
            this._updateIndex(change),
        ]);
        this._options.onReplicationComplete &&
            this._options.onReplicationComplete(this);
        return true;
    }
    async saveSnapshot() {
        const snapshotData = this._oplog.toSnapshot();
        const values = await snapshotData.values;
        const buf = serialize(new Snapshot({
            id: snapshotData.id,
            heads: snapshotData.heads,
            size: BigInt(values.length),
            values: values,
        }));
        const snapshot = await this._saveFile(buf);
        await this._cache.set(this.snapshotPath, serialize(new CID({ hash: snapshot })));
        await waitForAsync(async () => (await this._cache
            .get(this.snapshotPath)
            .then((bytes) => bytes && deserialize(bytes, CID))) !== undefined, { delayInterval: 200, timeout: 10 * 1000 });
        logger.debug(`Saved snapshot: ${snapshot}`);
        return [snapshot];
    }
    async loadFromSnapshot() {
        if (this._options.onLoad) {
            await this._options.onLoad(this);
        }
        await this.sync([]);
        const snapshotCID = await this._cache
            .get(this.snapshotPath)
            .then((bytes) => bytes && deserialize(bytes, CID));
        if (snapshotCID) {
            const file = await this._loadFile(snapshotCID.hash);
            if (!file) {
                throw new Error("Missing snapshot");
            }
            const snapshotData = deserialize(file, Snapshot);
            const headEntries = new Array(snapshotData.heads.length);
            const valuesMap = new Map();
            for (const v of snapshotData.values) {
                valuesMap.set(v.hash, v);
            }
            for (const [i, hash] of snapshotData.heads.entries()) {
                headEntries[i] = valuesMap.get(hash);
            }
            // Fetch the entries
            // Timeout 1 sec to only load entries that are already fetched (in order to not get stuck at loading)
            if (snapshotData) {
                this._oplog = await Log.fromEntry(this._store, this.identity, headEntries, {
                    sortFn: this._options.sortFn,
                    length: -1,
                    timeout: 1000,
                    onFetched: this._onLoadProgress.bind(this),
                });
                await this._updateIndex({
                    added: await this._oplog.values.toArray(),
                    removed: [],
                });
                this._options.onReplicationComplete &&
                    this._options.onReplicationComplete(this);
            }
            this._options.onReady && this._options.onReady(this);
        }
        else {
            throw new Error(`Snapshot for ${this.id} not found!`);
        }
        this._loaded = true;
        return this;
    }
    async _updateIndex(change) {
        // TODO add better error handling
        try {
            if (this._onUpdate) {
                await this._onUpdate(change);
            }
        }
        catch (error) {
            if (error instanceof AccessError) {
                // fail silently for now
                logger.info("Could not _onUpdate due to AccessError");
            }
            else {
                throw error;
            }
        }
        try {
            if (this._onUpdateOption) {
                await this._onUpdateOption(change);
            }
        }
        catch (error) {
            if (error instanceof AccessError) {
                // fail silently for now
                logger.info("Could not _onUpdateOption due to AccessError");
            }
            else {
                throw error;
            }
        }
    }
    /* Loading progress callback */
    _onLoadProgress(entry) {
        this._options.onLoadProgress && this._options.onLoadProgress(this, entry);
    }
    clone() {
        return deserialize(serialize(this), this.constructor);
    }
};
__decorate([
    field({ type: "u32" }),
    __metadata("design:type", Number)
], Store.prototype, "_storeIndex", void 0);
Store = __decorate([
    variant(0),
    __metadata("design:paramtypes", [Object])
], Store);
export { Store };
//# sourceMappingURL=store.js.map